﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Numerics;

namespace ConsoleApp8
{
	class Program
	{



		/**
		 * <summary>
		 * Entry point of the application
		 * </summary>
		 *
		 * <param name="args">command line parameters</param>
		 */
		public static void Main(string[] args)
		{
			// The first iterations can be slower, based on cpu (need to get to higher frequency if not set in performance mode)
			// and memory allocation with intelligent garbage collector make it faster because it re-use already allocated memory.
			for (int i = 0; i < 10; i++) {
				// clean up look up table calculated in previous computation to have fair time calculation
				_PowerLookUpTable = new Dictionary<int, Dictionary<int, System.Numerics.BigInteger>>();
				_PerformCalculation();
			}
			Console.WriteLine("Press any key to exit...");
			Console.ReadKey();
		}

		/**
		 * <summary>
		 * Function that perform the algorithm
		 * </summary>
		 */
		static private void _PerformCalculation()
		{
			System.Diagnostics.Stopwatch s1 = new System.Diagnostics.Stopwatch();
			s1.Start();

			long distinctSolution = 4000000;          // number of distinct solution we want to reach
			
			// googling around I found this interesting consideration:
			// 
			// Our problem is to solve: 1/x+1/y = 1/n. With the conditions: x,y,n are positive integer. 
			// We are going to try to simplify this equation.
			// To soddisfy the equation the variable "x" and "y" must be bigger than n. With this in mind we can substitute "x" by "n+a" and "y" by "n+b" (where "a" and "b" are positive integer) which lead to:
			//  1      1      1
			// ---- + ----  = ---
			// (n+a)  (n+b)    n
			//	We perform the first sum
			// (n+b)+(n+a)     1
			// -----------  = ---
			// (n+a)*(n+b)     n
			// we bring the denominator in the numerator
			// n*((n+b) + (n+a)) = (n+b)*(n+a)
			// we perform the multiplication
			// n^2 + nb + n^2 + na = n^2 + nb + na + ab
			// which finally lead to:
			// n^2 = ab
			// now we must keep in mind that our problem require "distinct" solution -> "a" and "b" cannot be swapped and "a" must be different from "b"
			// we therefore need to find X number of a and b combination ( variable limit )
			long limit = 2 * distinctSolution - 1;

			// we can use the prime factorization to find the number of divisors, the other number will be created using exponent applied on prime numbers
			// @see https://en.wikipedia.org/wiki/Prime_number#Unique_factorization
			int[] primeNumbers = SieveOfEratosthenes(50); // we limit the sieves to 50 give us 15 prime number which are enough to solve our problem

			//variable to hold the exponents used for our calculation
			int[] primeNumberExponents = new int[primeNumbers.Length];

			//find the absolute max result (were we multiply all the prime factor), from this we will try to find if there is a smaller solution
			System.Numerics.BigInteger result = 1;
			for (int i = 0; i < primeNumbers.Length; i++) {
				result *= primeNumbers[i];
				_PowerLookUpTable.Add(primeNumbers[i], new Dictionary<int, System.Numerics.BigInteger>());
			}

			int counter = 1;
			// start the search for the minimum
			while (true) {
				// function that calculate the primeNumberExponent of the prime number 2 to reach the result, the calculation take in consideration the other prime number (3,5...etc) exponent
				CalculatePrimeNumberTwoExponent(limit, primeNumberExponents);

				// when we have a candidate number where the number of the exponent of prime number 2 is lower than the number of the exponent of prime number 3
				// if not, we need to increment the exponent of a larger prime factor and then set all smaller prime factors’ exponents to the same number 
				if (primeNumberExponents[0] < primeNumberExponents[1]) {					
					counter++;
				}
				else {
					// The number of twos is larger than the number of threes, so this is a good solution 
					// if the result is minor than the previous one then we update the result
					System.Numerics.BigInteger number = PrimeNumberMul(primeNumbers, primeNumberExponents, result);
					if (number < result) {
						result = number;
					}
					counter = 1;
				}

				// if we have reached the limit of our prime number, we can stop the search
				if (counter >= primeNumberExponents.Length) {
					break;
				}

				// increase the current prime number exponent and set the smaller prime numbers exponent to the current exponent value
				primeNumberExponents[counter]++;
				for (int i = 0; i < counter; i++) {
					primeNumberExponents[i] = primeNumberExponents[counter];
				}
			}
			s1.Stop();

			Console.WriteLine("The first n with more than "+ ((limit + 1) / 2) + " solutions is " + result );
			Console.WriteLine("Solution took " + s1.ElapsedMilliseconds + " ms");
			System.Console.WriteLine("---------------------------");
		}

		/**
		 * <summary>
		 * Function that calculate multiplication obtained from a set of prime number with a specific exponent
		 * The function use _PowerLookUpTable static variable to record some calculation and reuse it when necessary
		 * </summary>
		 * 
		 * <param name="primeNumbers">array of prime number to multiply</param>
		 * <param name="primeNumberExponents">array of the prime number exponent (the index reflect prime number array)</param>
		 * <param name="result">previous calculated result</param>
		 * <return>the multiplication or "previous result" + 1 in case the result is bigger than previous result</return>
		 */
		static public System.Numerics.BigInteger PrimeNumberMul(int[] primeNumbers, int[] primeNumberExponents, System.Numerics.BigInteger result)
		{
			System.Numerics.BigInteger number = 1;
			System.Numerics.BigInteger tmp = 0;
			for (int i = 0; i < primeNumberExponents.Length; i++) {

				// break on the first exponent set to 0 because all the others will be 0
				if (primeNumberExponents[i] == 0) return number;

				// we must perform this operation
				// number *= System.Numerics.BigInteger.Pow(primeNumber[i], primeNumberExponents[i]);
				// pow is an expensive calculation, we keep track of the previous power calculated to make the function faster
				if (_PowerLookUpTable[primeNumbers[i]].ContainsKey(primeNumberExponents[i])) {
					number *= _PowerLookUpTable[primeNumbers[i]][primeNumberExponents[i]];
				}
				else {
					if (i == 0) { // if prime number is 2, the operation is a simply bit shift
						tmp = _BigIntegerOne << primeNumberExponents[i];
					}
					else {
						tmp = System.Numerics.BigInteger.Pow(primeNumbers[i], primeNumberExponents[i]);
					}
					number *= tmp;
					_PowerLookUpTable[primeNumbers[i]].Add(primeNumberExponents[i], tmp);
				}
				
				// in case the result is bigger than previous result we can exit with result +1 to make the calculation faster, we dont need the real calculation
				if (result < number) {
					return result+1;
				}
			}
			return number;
		}

		/**
		 * <summary>
		 * Function that calculate the primeNumberExponent of the prime number 2 to reach the result, this calculation is derived from the other prime number (3,5...etc) exponent
		 * </summary>
		 * 
		 * <param name="limit">result to reach</param>
		 * <param name="primeNumberExponents">array of prime number exponent (the index reflect prime number array)</param>
		 */
		static public void CalculatePrimeNumberTwoExponent(long limit, int[] primeNumberExponents)
		{
			primeNumberExponents[0] = 0;
			long divisors = 1;

			for (int i = 0; i < primeNumberExponents.Length; i++) {
				divisors *= 2 * primeNumberExponents[i] + 1;
			}
			primeNumberExponents[0] = (int)(limit / divisors - 1) / 2;

			while (divisors * (2 * primeNumberExponents[0] + 1) < limit) {
				primeNumberExponents[0]++;
			}
		}

		/**
		 * <summary>
		 * Function, based on Sieve of Eratosthenes (https://en.wikipedia.org/wiki/Sieve_of_Eratosthenes) 
		 * to calculate in a fast way the prime number up to a certain limit
		 * 
		 * Eratosthenes' method:
		 * 
		 * 1) Create a list of consecutive integers from 2 through n: (2, 3, 4, ..., n).
		 * 2) Initially, let p equal 2, the smallest prime number.
		 * 3) Enumerate the multiples of p by counting in increments of p from 2p to n, and mark them in the list (these will be 2p, 3p, 4p, ...; the p itself should not be marked).
		 * 4) Find the first number greater than p in the list that is not marked. If there was no such number, stop. Otherwise, let p now equal this new number (which is the next prime), and repeat from step 3.
		 * 5) When the algorithm terminates, the numbers remaining not marked in the list are all the primes below n.
		 * 
		 * </summary>
		 * 
		 * <param name="limit">prime number limit</param>
		 * <return>A list of prime number up to limit</return>
		 */
		static public int[] SieveOfEratosthenes(int limit)
		{
			// Create a list of integers (up to sieveBound) all marked as prime
			// in this list we skip 0,1, 2 and all multiple of 2 (skipping index 0 -> the formula is 2 * i + 1. Example:
			// primeNumber[0] -> 2 (2*0 +1)
			// primeNumber[1] -> 3 (2*1 +1)
			// primeNumber[2] -> 5 (2*2 +1)
			// primeNumber[3] -> 9 (2*3 +1)
			// this notation save some memory and computation time because we skip the first 2^n iteration and we dont memorize them
			// we use BitArray because it's limited in memory and has a rellay fast read / write performance
			int ListBound = (int)(limit - 1) / 2;
			System.Collections.BitArray bitArray = new System.Collections.BitArray(ListBound + 1, true);
			//Iterate through the list, till the square root of the limit. We use the square root because of the meaning of the array
			for (int i = 1; i <= ((int)Math.Sqrt(limit) - 1) / 2; i++) {
				// if marked as prime
				if (bitArray[i]) {
					// now we erase from prime number all of the number multiples of the current
					for (int j = i * 2 * (i + 1); j <= ListBound; j += 2 * i + 1) {
						bitArray[j] = false;
					}
				}
			}

			// now we have our prime number list but in a very particulare format...let's convert it in a list of number accessible directly
			// to calculate the number of prime number https://en.wikipedia.org/wiki/Prime-counting_function			
			List<int> primeNumberList = new List<int>();
			int cnt = 0;
			primeNumberList.Add(2);
			for (int i = 1; i <= ListBound; i++) {
				if (bitArray.Get(i)) {
					int vl = 2 * i + 1;
					primeNumberList.Add(vl);
					if (vl == 0) break;
				}
			}
			return primeNumberList.ToArray();
		}


		/**
		 * look up table to make power faster if already calculated
		 */
		static Dictionary<int, Dictionary<int, System.Numerics.BigInteger>> _PowerLookUpTable = new Dictionary<int, Dictionary<int, System.Numerics.BigInteger>>();

		/**
		 * 1 expressed with BigInteger type
		 */
		static readonly System.Numerics.BigInteger _BigIntegerOne = 1;
	}
}
